#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __spreadArray = (this && this.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.graphqlTypescriptCodegen = void 0;
var lodash_1 = __importDefault(require("lodash"));
var fs_1 = __importDefault(require("fs"));
var glob_1 = __importDefault(require("glob"));
var path_1 = __importDefault(require("path"));
var superagent_1 = __importDefault(require("superagent"));
var graphql_1 = require("graphql");
var typescript_1 = require("./typescript");
var util_1 = require("./util");
var chalk_1 = __importDefault(require("chalk"));
var prompts_1 = __importDefault(require("prompts"));
var gqlmin_1 = __importDefault(require("gqlmin"));
function writeLogFunc(quiet) {
    if (quiet)
        return lodash_1.default.noop;
    return function (message) {
        var ts = new Date();
        var h = lodash_1.default.padStart("" + ts.getHours(), 2, "0");
        var m = lodash_1.default.padStart("" + ts.getMinutes(), 2, "0");
        var s = lodash_1.default.padStart("" + ts.getSeconds(), 2, "0");
        var mi = lodash_1.default.padStart("" + ts.getMilliseconds(), 3, "0");
        console.log("[" + chalk_1.default.gray(h + ":" + m + ":" + s + "." + mi) + "] " + message);
    };
}
function mapPrefix(prefixMap, v) {
    for (var key in prefixMap) {
        var val = prefixMap[key] || "";
        v = v.replace(key, val);
    }
    return v;
}
function exportInterfaceOrType(name, type) {
    if (type.startsWith("{")) {
        return "export interface " + name + " " + type;
    }
    else {
        return "export type " + name + " = " + type + ";";
    }
}
function generate(schema, doc, imports) {
    var _a = typescript_1.resolveTypesSorted(schema, doc, imports), usedNamedTypesSorted = _a.usedNamedTypesSorted, fragmentsSorted = _a.fragmentsSorted, operationsSorted = _a.operationsSorted, fragments = _a.fragments, fragmentDeps = _a.fragmentDeps;
    var types = "";
    var files = [];
    if (imports.embedImports) {
        imports.loadedImportsMap = {};
    }
    else {
        for (var _i = 0, _b = lodash_1.default.keys(imports.loadedImportsMap).sort(); _i < _b.length; _i++) {
            var impName = _b[_i];
            var imp = util_1.nonNull(imports.loadedImportsMap[impName], "import");
            types += "import { " + imp.name + "Fragment } from \"" + mapPrefix(imports.prefixMap, imp.path) + "/types\";\n\n";
        }
    }
    types += "export type ArbitraryObjectType = any;\n\n";
    for (var _c = 0, usedNamedTypesSorted_1 = usedNamedTypesSorted; _c < usedNamedTypesSorted_1.length; _c++) {
        var nt = usedNamedTypesSorted_1[_c];
        var t = schema.getType(nt.name);
        if (graphql_1.isEnumType(t)) {
            types += "export type " + nt.name + " = " + typescript_1.enumToString(t) + ";\n\n";
        }
        else if (graphql_1.isInputObjectType(t)) {
            var tt = typescript_1.convertType(t, undefined, fragments, imports.loadedImportsMap, true, true);
            tt.nullable = false; // hack: top level types are not nullable
            types += exportInterfaceOrType(nt.name, typescript_1.typeToString(tt, 0, true)) + "\n\n";
        }
        else {
            // must be a scalar type, skip it
        }
    }
    for (var _d = 0, fragmentsSorted_1 = fragmentsSorted; _d < fragmentsSorted_1.length; _d++) {
        var frag = fragmentsSorted_1[_d];
        types += exportInterfaceOrType(frag.name + "Fragment", typescript_1.typeToString(frag.value.type)) + "\n\n";
        files.push({
            name: "fragments/" + frag.name + ".ts",
            text: "export default " + JSON.stringify(gqlmin_1.default(graphql_1.print(frag.value.node)) + "\n") + ";",
        });
    }
    for (var _e = 0, operationsSorted_1 = operationsSorted; _e < operationsSorted_1.length; _e++) {
        var op = operationsSorted_1[_e];
        types += exportInterfaceOrType(op.name, typescript_1.typeToString(op.value.operation, 0, false)) + "\n\n";
        types += exportInterfaceOrType(op.name + "Variables", typescript_1.typeToString(op.value.variables, 0, true)) + "\n\n";
        types += "export interface " + op.name + "Meta {\n";
        types += "  __opType: " + op.name + ", __opVariablesType: " + op.name + "Variables, __tag: 'graphql-operation'\n";
        types += "}\n\n";
    }
    for (var _f = 0, operationsSorted_2 = operationsSorted; _f < operationsSorted_2.length; _f++) {
        var op = operationsSorted_2[_f];
        var fragsFromOp = {};
        typescript_1.extractFragmentDeps(op.value.node.selectionSet, fragsFromOp);
        var fragNames = lodash_1.default.keys(fragsFromOp).sort();
        var allDepImports = {};
        var allDepNames = {};
        for (var _g = 0, fragNames_1 = fragNames; _g < fragNames_1.length; _g++) {
            var fragName = fragNames_1[_g];
            var deps = {};
            typescript_1.addFragmentDepsRecursive(deps, [fragName], fragmentDeps, imports.loadedImportsMap);
            for (var _h = 0, _j = lodash_1.default.keys(deps).sort(); _h < _j.length; _h++) {
                var dep = _j[_h];
                var _k = lodash_1.default.split(dep, ":", 2), path_2 = _k[0], name_1 = _k[1];
                allDepNames[name_1] = true;
                allDepImports["import " + name_1 + " from \"" + mapPrefix(imports.prefixMap, path_2) + "/fragments/" + name_1 + "\";"] = true;
            }
        }
        var depNames = lodash_1.default.keys(allDepNames).sort();
        var depImports = lodash_1.default.join(__spreadArray(__spreadArray([], lodash_1.default.keys(allDepImports).sort()), ["import { " + op.name + "Meta } from \"../types\";"]), "\n");
        var strings = lodash_1.default.join(__spreadArray(__spreadArray([], depNames), [JSON.stringify(gqlmin_1.default(graphql_1.print(op.value.node)))]), " + ");
        files.push({
            name: "operations/" + op.name + ".ts",
            text: depImports + "\nexport default (" + strings + ") as unknown as " + op.name + "Meta;",
        });
    }
    return {
        types: types,
        files: files,
    };
}
function writeFileIfChanged(writeLog, path, content) {
    try {
        var data = fs_1.default.readFileSync(path, "utf8");
        if (data === content) {
            writeLog("no changes in file \"" + path + "\"");
            return;
        }
    }
    catch (_a) { }
    fs_1.default.writeFileSync(path, content, "utf8");
    writeLog("writing data to file \"" + path + "\"");
}
function addTokenMaybe(token) {
    return function (req) {
        if (token)
            req.set("Authorization", "Bearer " + token);
    };
}
function fetchSchema(writeLog, pathOrURL, token) {
    return __awaiter(this, void 0, void 0, function () {
        var introspectionQuery, resp, data;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!lodash_1.default.startsWith(pathOrURL, "https://")) return [3 /*break*/, 2];
                    writeLog("fetching schema via introspection query from \"" + pathOrURL + "\"");
                    introspectionQuery = graphql_1.getIntrospectionQuery({ descriptions: false, inputValueDeprecation: false });
                    return [4 /*yield*/, superagent_1.default
                            .post(pathOrURL)
                            .http2()
                            .accept("json")
                            .use(addTokenMaybe(token))
                            .send({ query: introspectionQuery })];
                case 1:
                    resp = _a.sent();
                    if (resp.body.errors)
                        throw new Error("failed fetching schema: " + resp.body.errors);
                    return [2 /*return*/, graphql_1.buildClientSchema(resp.body.data)];
                case 2:
                    writeLog("loading schema from file \"" + pathOrURL + "\"");
                    data = fs_1.default.readFileSync(pathOrURL, "utf8");
                    return [2 /*return*/, graphql_1.buildSchema(data)];
            }
        });
    });
}
function graphqlTypescriptCodegen(config) {
    return __awaiter(this, void 0, void 0, function () {
        var writeLog, tsPattern, tsFilePaths, tsFilePathsMap, graphqlPattern, graphqlFilePaths, sources, allImports, _i, graphqlFilePaths_1, input, src, doc, schema, _a, imports, additionalFragments, _b, _c, imp, deps, dep, _d, name_2, frag, data, _e, _f, dir, _g, _h, f, unusedFiles, _j, unusedFiles_1, f, confirmed, _k, unusedFiles_2, f, _l, unusedFiles_3, f;
        return __generator(this, function (_m) {
            switch (_m.label) {
                case 0:
                    writeLog = writeLogFunc(config.quiet);
                    tsPattern = path_1.default.join(config.directory, "**", "*.ts");
                    tsFilePaths = glob_1.default.sync(tsPattern);
                    tsFilePathsMap = lodash_1.default.fromPairs(lodash_1.default.map(tsFilePaths, function (p) { return [p, true]; }));
                    graphqlPattern = path_1.default.join(config.directory, "**", "*.graphql");
                    graphqlFilePaths = glob_1.default.sync(graphqlPattern);
                    sources = [];
                    allImports = [];
                    for (_i = 0, graphqlFilePaths_1 = graphqlFilePaths; _i < graphqlFilePaths_1.length; _i++) {
                        input = graphqlFilePaths_1[_i];
                        writeLog("loading graphql file \"" + input + "\"");
                        src = typescript_1.loadGraphQLFromFile(input);
                        allImports.push.apply(allImports, typescript_1.extractImportSpecs(src.body));
                        if (src)
                            sources.push(src);
                    }
                    doc = graphql_1.concatAST(lodash_1.default.map(sources, function (src) { return graphql_1.parse(src); }));
                    _a = config.schema;
                    if (_a) return [3 /*break*/, 2];
                    return [4 /*yield*/, fetchSchema(writeLog, config.schemaPathOrURL, config.token)];
                case 1:
                    _a = (_m.sent());
                    _m.label = 2;
                case 2:
                    schema = _a;
                    imports = typescript_1.loadImports(allImports, config.includes, schema, config.embedImports);
                    additionalFragments = {};
                    for (_b = 0, _c = imports.loadedImports; _b < _c.length; _b++) {
                        imp = _c[_b];
                        deps = {};
                        typescript_1.addFragmentDepsRecursive(deps, [imp.name], imp.context.fragmentDeps, {});
                        for (dep in deps) {
                            _d = lodash_1.default.split(dep, ":", 2), name_2 = _d[1];
                            frag = util_1.nonNull(imp.context.fragments[name_2], "fragment: " + name_2);
                            additionalFragments[name_2] = {
                                kind: "Document",
                                definitions: [frag.node],
                            };
                        }
                    }
                    typescript_1.validateDocument(schema, graphql_1.concatAST(__spreadArray([doc], lodash_1.default.compact(lodash_1.default.values(additionalFragments)))), config.allowUnusedFragments);
                    data = generate(schema, doc, imports);
                    for (_e = 0, _f = [path_1.default.join(config.directory, "fragments"), path_1.default.join(config.directory, "operations")]; _e < _f.length; _e++) {
                        dir = _f[_e];
                        fs_1.default.mkdirSync(dir, { recursive: true });
                    }
                    writeFileIfChanged(writeLog, path_1.default.join(config.directory, "types.ts"), data.types);
                    delete tsFilePathsMap[config.directory + "/types.ts"];
                    for (_g = 0, _h = data.files; _g < _h.length; _g++) {
                        f = _h[_g];
                        writeFileIfChanged(writeLog, config.directory + "/" + f.name, f.text);
                        delete tsFilePathsMap[config.directory + "/" + f.name];
                    }
                    unusedFiles = lodash_1.default.keys(tsFilePathsMap).sort();
                    if (!(unusedFiles.length > 0)) return [3 /*break*/, 5];
                    if (!config.quiet) return [3 /*break*/, 3];
                    for (_j = 0, unusedFiles_1 = unusedFiles; _j < unusedFiles_1.length; _j++) {
                        f = unusedFiles_1[_j];
                        fs_1.default.renameSync(f, f + ".unused");
                    }
                    return [3 /*break*/, 5];
                case 3:
                    writeLog("the following files are no longer used:\n" + lodash_1.default.join(unusedFiles, "\n"));
                    return [4 /*yield*/, prompts_1.default({
                            type: "confirm",
                            name: "confirmed",
                            message: "delete them?",
                            initial: false,
                        })];
                case 4:
                    confirmed = (_m.sent()).confirmed;
                    if (confirmed) {
                        for (_k = 0, unusedFiles_2 = unusedFiles; _k < unusedFiles_2.length; _k++) {
                            f = unusedFiles_2[_k];
                            writeLog("deleting \"" + f + "\"");
                            fs_1.default.unlinkSync(f);
                        }
                    }
                    else {
                        for (_l = 0, unusedFiles_3 = unusedFiles; _l < unusedFiles_3.length; _l++) {
                            f = unusedFiles_3[_l];
                            writeLog("renaming \"" + f + "\" to \"" + f + ".unused\"");
                            fs_1.default.renameSync(f, f + ".unused");
                        }
                    }
                    _m.label = 5;
                case 5: return [2 /*return*/];
            }
        });
    });
}
exports.graphqlTypescriptCodegen = graphqlTypescriptCodegen;
